_ = require 'underscore'
# {events} = modules


# ---dependent functions (of CCNode)---
# ... addChild (essential)
# ... setVisible (for implement show prop)


# reserved at template
reservedProps = [
	'type',
	'bind',
	'template'
	'children'
	'repeat'
	'show'
#	'id'
	'priority'
]


to_arr = (arg) ->
	if _.isString arg
		arg.split(' ').map (c) -> parseFloat c
	else
		arg


randomId = ->
	Math.floor(Math.random() * 9007199254740992)


# return a function that return evaluated
makeFunc = (str) ->
	contents = [
		'for (var k in scope) {'
		'  this[k] = scope[k]'
		'}'
		"return (#{str});"
	]
	fn = new Function ['scope'], contents
	(scope) ->
		fn.call({}, scope)


# return a function that return interpolated str
interpolate = (str) ->
	resArr = []
	recent = 0
	str.replace /{{([^}]+)}}/g, (match, p1, offset, string) ->
		if recent < offset 
			resArr.push str.substr recent, offset - recent
		recent = offset + match.length
		resArr.push makeFunc(p1)
		''

	if recent < str.length
		resArr.push str.substr recent, str.length - recent

	(scope) ->
		something = null
		for mola in resArr
			if typeof mola == 'string'
				ala = mola
			else if typeof mola == 'function'
				ala = mola scope
			else
				throw 'unrecognized type of mola : ' + typeof mola

			unless something?
				something = ala
			else
				something += ala

		something

innerApply_r = ($scope) ->
	changed = false

	for w in $scope.$$watches
		cur = w.targetFunc.call($scope);
		cur_str = JSON.stringify(cur)
		if w.prev_str != cur_str
			w.listener.call($scope, cur, w.prev)
			w.prev = _.clone cur
			w.prev_str = cur_str
			changed = true

	_.reduce $scope.$$childScopes, ((memo, s) -> innerApply_r(s) or memo), changed
	
genScope = (parentScope) ->
	Scope = ->
	Scope.prototype = parentScope

	$scope = new Scope()
	$scope.$$childScopes = []
	$scope.$$watches = []

	is_root_scope = true if parentScope == null

	if is_root_scope
		$scope.$$rootScope = $scope
#		$scope.$$elements = {}
	else
		parentScope.$$childScopes.push scope
		scope.$$rootScope = parentScope.$$rootScope

	$scope.$$watch = (targetFunc, listener) ->
		$scope.$$watches.push
			targetFunc: targetFunc
			listener: listener
			prev: null
			prev_str: ''
	 	
	$scope.$apply = (fn) ->
		fn?()

		loop
			break unless innerApply_r $scope
		null

	$scope

module.exports = (dicts...) ->
	# make a appropriate drwable object dictionary...
	# compose dictionaries to one dictionary
	if dicts == null or dicts.length == 0
		dict = []
	else
		dict = dicts[0]
		dict = _.extend dict, dict[i] for i in [1...dicts.length]
	

	pub = {}

	pub.include = (newDict) ->
		dict = _.extend dict, newDict

	# traverse node
	# if node has controller... create scope
	createScope_r = (template, parentScope, controllers) ->
		is_root_scope = true unless parentScope
		node_has_scope = true if template.controller

		throw 'root node must have a controller' if is_root_scope and not node_has_scope

		if node_has_scope
			scope = genScope parentScope
			ctrl = controllers[template.controller]
			ctrl scope
			scope.$$controllerName = template.controller
		else
			scope = parentScope

		if template.children?
			for child in template.children
				createScope_r child, scope, controllers

	createNode_r = (template, scope, option) ->
		node_has_scope = true if template.controller? or option.type == 'repeated'

		def = dict[template.type]
		throw "unsupported type : #{template.type}" unless def?

		# create a node
		unless def.cons_args?
			node = do d.cons
		else
			initArgs = def.cons_args.map (arg) -> interpolate(template[arg])(scope)
			node = d.cons initArgs...

		if template.show?
			fn = interpolate template.show
			scope.$$watch fn, (res) ->
				if res == true or res == 'true'
					node.setVisible true
				else
					node.setVisible false

		if template.repeat?
			[itr,_in_,container] = template.repeat.bind.split(' ')
			throw 'repeat usage wrong... [IDENT \'in\' IDENT]' if _in_ != 'in'
			
			src = scope[container]
			throw "no #{container} in the scope" unless src?

			# don't need to call interpolate function because it already interpolated...
			fn = makeFunc template.src
			scope.$$watch fn, (list, prev_list) ->
				prev_list = [] if prev_list is null

				deleted_list = []
				for prev_item in prev_list
					unless _.every list, ((item) -> prev_item.$id != item.$id)
						deleted_list.push prev_item

				deleted_list.forEach (item) ->
#					scope.$$childScopes = _.without scope.$$childScopes, item.$$scope
					action1 = cc.Spawn.create(cc.ScaleBy.create(1, 1), cc.FadeOut.create(1))
					action2 = cc.CallFunc.create ->
						item.removeFromParent()
					item.runAction cc.Sequence.create(action1, action2)

				new_list = []
				for item in list
					if _.every prev_list, ((prev_item) -> prev_item.$id != item.$id)
						new_list.push item

				childTemplate = template.repeat.template
				new_list.forEach (item) ->
					childScope = genScope scope
					childScope[itr] = item
					childNode = createNode_r childTemplate, childScope
					node.addChild childNode, (childTemplate.priority or 0)

		if template.children?
			template.children.forEach (child, index) ->
				if child.controller?
					childScope = _.find scope.$$childScopes, (s) ->
						s.$$controllerName == child.controller
					throw "can\'t find @{child.controller} controller" if childScope == null	
				childNode = createNode_r child, childScope
				node.addChild childNode, (child.priority or 0)

		managedProps = []
		for name of template
			fnMap[name] 
			prop = template[name]
			if not _.contains reservedProps, name
				managedProps.push name

		while def?
			managedProps.forEach (name, idx) ->
				prop = def[name]
				if typeof prop == 'function'
					# register
					scope.$$watch interpolate(template[name]), (value) ->
						prop.call node, value

					delete managedProps[idx]

			managedProps = _.compact managedProps
			def = dict[d.inherit]

		node

	pub.makeDom = (template, controllers) ->
		rootScope = createScope_r template, null, controllers
		rootNode = createNode_r template, rootScope
		rootScope.$apply -> # for initiation
		rootNode

	pub